;; Generated package description from chess.el  -*- no-byte-compile: t -*-
(define-package "chess" "2.0.5.0.20201209.142246" "Play chess in GNU Emacs" '((cl-lib "0.5")) :url "https://elpa.gnu.org/packages/chess.html" :authors '(("John Wiegley" . "johnw@gnu.org")) :maintainer '("Mario Lang" . "mlang@delysid.org") :keywords '("games"))
